@@ -4,8 +4,7 @@ import Control.Applicative
 import Control.Arrow (second)
 import Control.Monad
 import Data.Typeable
-import Test.QuickCheck (Gen, Property, Testable)
-import Test.QuickCheck qualified as QC
+import Test.QuickCheck hiding (generate)
 import Test.QuickCheck.DynamicLogic.CanGenerate
 import Test.QuickCheck.DynamicLogic.Quantify
 import Test.QuickCheck.DynamicLogic.SmartShrinking
@@ -340,8 +339,7 @@ restrictedPolar (ActionWithPolarity a _) = restricted a
 
 -- | Simplest "execution" function for `DynFormula`.
 -- Turns a given a `DynFormula` paired with an interpreter function to produce some result from an
-
---- `Actions` sequence into a proper `Property` than can then be run by QuickCheck.
+-- `Actions` sequence into a proper `Property` that can be run by QuickCheck.
 forAllScripts
   :: (DynLogicModel s, Testable a)
   => DynFormula s
@@ -361,8 +359,8 @@ forAllUniqueScripts s f k =
     let d = unDynFormula f sz
         n = unsafeNextVarIndex $ vars s
      in case generate chooseUniqueNextStep d n s 500 of
-          Nothing -> QC.counterexample "Generating Non-unique script in forAllUniqueScripts" False
-          Just test -> validDLTest d test . applyMonitoring d test . QC.property $ k (scriptFromDL test)
+          Nothing -> counterexample "Generating Non-unique script in forAllUniqueScripts" False
+          Just test -> validDLTest test . applyMonitoring d test . property $ k (scriptFromDL test)
 
 -- | Creates a `Property` from `DynFormula` with some specialised isomorphism for shrinking purpose.
 forAllMappedScripts
@@ -375,22 +373,37 @@ forAllMappedScripts
 forAllMappedScripts to from f k =
   QC.withSize $ \n ->
     let d = unDynFormula f n
-     in QC.forAllShrinkBlind
-          (QC.Smart 0 <$> QC.sized ((from <$>) . generateDLTest d))
+     in forAllShrinkBlind
+          (Smart 0 <$> sized ((from <$>) . generateDLTest d))
           (shrinkSmart ((from <$>) . shrinkDLTest d . to))
-          $ \(QC.Smart _ script) ->
+          $ \(Smart _ script) ->
             withDLScript d k (to script)
 
 withDLScript :: (DynLogicModel s, Testable a) => DynLogic s -> (Actions s -> a) -> DynLogicTest s -> Property
 withDLScript d k test =
-  validDLTest d test . applyMonitoring d test . QC.property $ k (scriptFromDL test)
+  validDLTest test . applyMonitoring d test . property $ k (scriptFromDL test)
 
 withDLScriptPrefix :: (DynLogicModel s, Testable a) => DynFormula s -> (Actions s -> a) -> DynLogicTest s -> Property
 withDLScriptPrefix f k test =
   QC.withSize $ \n ->
     let d = unDynFormula f n
         test' = unfailDLTest d test
-     in validDLTest d test' . applyMonitoring d test' . QC.property $ k (scriptFromDL test')
+     in validDLTest test' . applyMonitoring d test' . property $ k (scriptFromDL test')
+
+-- | Validate generated test case.
+--
+-- Test case generation does not always produce a valid test case. In
+-- some cases, we did not find a suitable test case matching some
+-- `DynFormula` and we are `Stuck`, hence we want to discard the test
+-- case and start over ; in other cases we found a genuine issue with
+-- the formula leading to the impossibility of producing a valid test
+-- case.
+validDLTest :: StateModel s => DynLogicTest s -> Property -> Property
+validDLTest test prop =
+  case test of
+    DLScript{} -> counterexample (show test) prop
+    Stuck{} -> property Discard
+    _other -> counterexample (show test) False
 
 generateDLTest :: DynLogicModel s => DynLogic s -> Int -> Gen (DynLogicTest s)
 generateDLTest d size = generate chooseNextStep d 0 (initialStateFor d) size
@@ -503,7 +516,7 @@ nextSteps' gen (ForAll q f) = do
 nextSteps' gen (Monitor _f d) = nextSteps' gen d
 
 chooseOneOf :: [(Double, a)] -> Gen a
-chooseOneOf steps = QC.frequency [(round (w / never), return s) | (w, s) <- steps]
+chooseOneOf steps = frequency [(round (w / never), return s) | (w, s) <- steps]
 
 never :: Double
 never = 1.0e-9
@@ -573,7 +586,7 @@ keepTryingUntil :: Int -> Gen a -> (a -> Bool) -> Gen (Maybe a)
 keepTryingUntil 0 _ _ = return Nothing
 keepTryingUntil n g p = do
   x <- g
-  if p x then return $ Just x else QC.scale (+ 1) $ keepTryingUntil (n - 1) g p
+  if p x then return $ Just x else scale (+ 1) $ keepTryingUntil (n - 1) g p
 
 shrinkDLTest :: DynLogicModel s => DynLogic s -> DynLogicTest s -> [DynLogicTest s]
 shrinkDLTest _ (Looping _) = []
@@ -697,7 +710,7 @@ demonicAlt ds = foldr1 (Alt Demonic) ds
 
 propPruningGeneratedScriptIsNoop :: DynLogicModel s => DynLogic s -> Property
 propPruningGeneratedScriptIsNoop d =
-  QC.forAll (QC.sized $ \n -> QC.choose (1, max 1 n) >>= generateDLTest d) $ \test ->
+  forAll (sized $ \n -> choose (1, max 1 n) >>= generateDLTest d) $ \test ->
     let script = case test of
           BadPrecondition s _ _ -> s
           Looping s -> s
@@ -764,11 +777,6 @@ stuck (Weight w d) s = w < never || stuck d s
 stuck (ForAll _ _) _ = False
 stuck (Monitor _ d) s = stuck d s
 
-validDLTest :: StateModel s => DynLogic s -> DynLogicTest s -> Property -> Property
-validDLTest _ Stuck{} _ = False QC.==> False
-validDLTest _ test@DLScript{} p = QC.counterexample (show test) p
-validDLTest _ test _ = QC.counterexample (show test) False
-
 scriptFromDL :: DynLogicTest s -> Actions s
 scriptFromDL (DLScript s) = Actions $ sequenceSteps s
 scriptFromDL _ = Actions []
