@@ -1,5 +1,7 @@
 {-# LANGUAGE AllowAmbiguousTypes #-}
+{-# LANGUAGE FunctionalDependencies #-}
 {-# LANGUAGE QuantifiedConstraints #-}
+{-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE UndecidableInstances #-}
 
 -- | Model-Based Testing library for use with Haskell QuickCheck.
@@ -23,9 +25,9 @@ module Test.QuickCheck.StateModel (
   EnvEntry (..),
   pattern (:=?),
   Env,
-  Realized,
   Generic,
   IsPerformResult,
+  Options (..),
   monitorPost,
   counterexamplePost,
   stateAfter,
@@ -38,12 +40,14 @@ module Test.QuickCheck.StateModel (
   computePrecondition,
   computeArbitraryAction,
   computeShrinkAction,
+  generateActionsWithOptions,
+  shrinkActionsWithOptions,
+  defaultOptions,
+  moreActions,
 ) where
 
 import Control.Monad
-import Control.Monad.Identity
 import Control.Monad.Reader
-import Control.Monad.State
 import Control.Monad.Writer (WriterT, runWriterT, tell)
 import Data.Data
 import Data.Kind
@@ -52,8 +56,7 @@ import Data.Monoid (Endo (..))
 import Data.Set qualified as Set
 import Data.Void
 import GHC.Generics
-import Test.QuickCheck (Arbitrary, Gen, Property, Smart (..), counterexample, frequency, resize, shrinkList, sized, tabulate)
-import Test.QuickCheck qualified as QC
+import Test.QuickCheck as QC
 import Test.QuickCheck.DynamicLogic.SmartShrinking
 import Test.QuickCheck.Monadic
 import Test.QuickCheck.StateModel.Variables
@@ -107,13 +110,6 @@ class
   -- anything.
   data Action state a
 
-  -- | The type of errors that actions can throw. If this is defined as anything
-  -- other than `Void` `perform` is required to return `Either (Error state) a`
-  -- instead of `a`.
-  type Error state
-
-  type Error state = Void
-
   -- | Display name for `Action`.
   -- This is useful to provide sensible statistics about the distribution of `Action`s run
   -- when checking a property.
@@ -163,31 +159,6 @@ class
 
 deriving instance (forall a. Show (Action state a)) => Show (Any (Action state))
 
--- | The result required of `perform` depending on the `Error` type
--- of a state model. If there are no errors, `Error state = Void`, and
--- so we don't need to specify if the action failed or not.
-type family PerformResult e a where
-  PerformResult Void a = a
-  PerformResult e a = Either e a
-
-class IsPerformResult e a where
-  performResultToEither :: PerformResult e a -> Either e a
-
-instance {-# OVERLAPPING #-} IsPerformResult Void a where
-  performResultToEither = Right
-
-instance {-# OVERLAPPABLE #-} (PerformResult e a ~ Either e a) => IsPerformResult e a where
-  performResultToEither = id
-
--- TODO: maybe it makes sense to write
--- out a long list of these instances
-type family Realized (m :: Type -> Type) a :: Type
-type instance Realized IO a = a
-type instance Realized (StateT s m) a = Realized m a
-type instance Realized (ReaderT r m) a = Realized m a
-type instance Realized (WriterT w m) a = Realized m a
-type instance Realized Identity a = a
-
 newtype PostconditionM m a = PostconditionM {runPost :: WriterT (Endo Property, Endo Property) m a}
   deriving (Functor, Applicative, Monad)
 
@@ -211,7 +182,33 @@ monitorPost m = PostconditionM $ tell (Endo m, mempty)
 counterexamplePost :: Monad m => String -> PostconditionM m ()
 counterexamplePost c = PostconditionM $ tell (mempty, Endo $ counterexample c)
 
+-- | The result required of `perform` depending on the `Error` type.
+-- If there are no errors, `Error state = Void`, and
+-- so we don't need to specify if the action failed or not.
+type family PerformResult state (m :: Type -> Type) a where
+  PerformResult state m a = EitherIsh (Error state m) a
+
+type family EitherIsh e a where
+  EitherIsh Void a = a
+  EitherIsh e a = Either e a
+
+class IsPerformResult e a where
+  performResultToEither :: EitherIsh e a -> Either e a
+
+instance {-# OVERLAPPING #-} IsPerformResult Void a where
+  performResultToEither = Right
+
+instance {-# OVERLAPPABLE #-} (EitherIsh e a ~ Either e a) => IsPerformResult e a where
+  performResultToEither = id
+
 class (forall a. Show (Action state a), Monad m) => RunModel state m where
+  -- | The type of errors that actions can throw. If this is defined as anything
+  -- other than `Void` `perform` is required to return `Either (Error state) a`
+  -- instead of `a`.
+  type Error state m
+
+  type Error state m = Void
+
   -- | Perform an `Action` in some `state` in the `Monad` `m`.  This
   -- is the function that's used to exercise the actual stateful
   -- implementation, usually through various side-effects as permitted
@@ -222,57 +219,57 @@ class (forall a. Show (Action state a), Monad m) => RunModel state m where
   --
   -- The `Lookup` parameter provides an /environment/ to lookup `Var
   -- a` instances from previous steps.
-  perform :: Typeable a => state -> Action state a -> LookUp m -> m (PerformResult (Error state) (Realized m a))
+  perform :: Typeable a => state -> Action state a -> LookUp -> m (PerformResult state m a)
 
   -- | Postcondition on the `a` value produced at some step.
   -- The result is `assert`ed and will make the property fail should it be `False`. This is useful
   -- to check the implementation produces expected values.
-  postcondition :: (state, state) -> Action state a -> LookUp m -> Realized m a -> PostconditionM m Bool
+  postcondition :: (state, state) -> Action state a -> LookUp -> a -> PostconditionM m Bool
   postcondition _ _ _ _ = pure True
 
   -- | Postcondition on the result of running a _negative_ `Action`.
   -- The result is `assert`ed and will make the property fail should it be `False`. This is useful
   -- to check the implementation produces e.g. the expected errors or to check that the SUT hasn't
   -- been updated during the execution of the negative action.
-  postconditionOnFailure :: (state, state) -> Action state a -> LookUp m -> Either (Error state) (Realized m a) -> PostconditionM m Bool
+  postconditionOnFailure :: (state, state) -> Action state a -> LookUp -> Either (Error state m) a -> PostconditionM m Bool
   postconditionOnFailure _ _ _ _ = pure True
 
   -- | Allows the user to attach additional information to the `Property` at each step of the process.
   -- This function is given the full transition that's been executed, including the start and ending
   -- `state`, the `Action`, the current environment to `Lookup` and the value produced by `perform`
   -- while executing this step.
-  monitoring :: (state, state) -> Action state a -> LookUp m -> Either (Error state) (Realized m a) -> Property -> Property
+  monitoring :: (state, state) -> Action state a -> LookUp -> Either (Error state m) a -> Property -> Property
   monitoring _ _ _ _ prop = prop
 
   -- | Allows the user to attach additional information to the `Property` if a positive action fails.
-  monitoringFailure :: state -> Action state a -> LookUp m -> Error state -> Property -> Property
+  monitoringFailure :: state -> Action state a -> LookUp -> Error state m -> Property -> Property
   monitoringFailure _ _ _ _ prop = prop
 
-type LookUp m = forall a. Typeable a => Var a -> Realized m a
+type LookUp = forall a. Typeable a => Var a -> a
 
-type Env m = [EnvEntry m]
+type Env = [EnvEntry]
 
-data EnvEntry m where
-  (:==) :: Typeable a => Var a -> Realized m a -> EnvEntry m
+data EnvEntry where
+  (:==) :: Typeable a => Var a -> a -> EnvEntry
 
 infix 5 :==
 
-pattern (:=?) :: forall a m. Typeable a => Var a -> Realized m a -> EnvEntry m
+pattern (:=?) :: forall a. Typeable a => Var a -> a -> EnvEntry
 pattern v :=? val <- (viewAtType -> Just (v, val))
 
-viewAtType :: forall a m. Typeable a => EnvEntry m -> Maybe (Var a, Realized m a)
+viewAtType :: forall a. Typeable a => EnvEntry -> Maybe (Var a, a)
 viewAtType ((v :: Var b) :== val)
   | Just Refl <- eqT @a @b = Just (v, val)
   | otherwise = Nothing
 
-lookUpVarMaybe :: forall a m. Typeable a => Env m -> Var a -> Maybe (Realized m a)
+lookUpVarMaybe :: forall a. Typeable a => Env -> Var a -> Maybe a
 lookUpVarMaybe [] _ = Nothing
 lookUpVarMaybe (((v' :: Var b) :== a) : env) v =
   case eqT @a @b of
     Just Refl | v == v' -> Just a
     _ -> lookUpVarMaybe env v
 
-lookUpVar :: Typeable a => Env m -> Var a -> Realized m a
+lookUpVar :: Typeable a => Env -> Var a -> a
 lookUpVar env v = case lookUpVarMaybe env v of
   Nothing -> error $ "Variable " ++ show v ++ " is not bound at type " ++ show (typeRep v) ++ "!"
   Just a -> a
@@ -368,56 +365,100 @@ usedVariables (Actions as) = go initialAnnotatedState as
         <> go (computeNextState aState act var) steps
 
 instance forall state. StateModel state => Arbitrary (Actions state) where
-  arbitrary = do
-    (as, rejected) <- arbActions initialAnnotatedState 1
-    return $ Actions_ rejected (Smart 0 as)
-    where
-      arbActions :: Annotated state -> Int -> Gen ([Step state], [String])
-      arbActions s step = sized $ \n ->
-        let w = n `div` 2 + 1
-         in frequency
-              [ (1, return ([], []))
-              ,
-                ( w
-                , do
-                    (mact, rej) <- satisfyPrecondition
-                    case mact of
-                      Just (Some act@ActionWithPolarity{}) -> do
-                        let var = mkVar step
-                        (as, rejected) <- arbActions (computeNextState s act var) (step + 1)
-                        return ((var := act) : as, rej ++ rejected)
-                      Nothing ->
-                        return ([], [])
-                )
-              ]
-        where
-          satisfyPrecondition = sized $ \n -> go n (2 * n) [] -- idea copied from suchThatMaybe
-          go m n rej
-            | m > n = return (Nothing, rej)
-            | otherwise = do
-                a <- resize m $ computeArbitraryAction s
-                case a of
-                  Some act ->
-                    if computePrecondition s act
-                      then return (Just (Some act), rej)
-                      else go (m + 1) n (actionName (polarAction act) : rej)
+  arbitrary = generateActionsWithOptions defaultOptions
+  shrink = shrinkActionsWithOptions defaultOptions
 
-  shrink (Actions_ rs as) =
-    map (Actions_ rs) (shrinkSmart (map (prune . map fst) . concatMap customActionsShrinker . shrinkList shrinker . withStates) as)
-    where
-      shrinker :: (Step state, Annotated state) -> [(Step state, Annotated state)]
-      shrinker (v := act, s) = [(unsafeCoerceVar v := act', s) | Some act'@ActionWithPolarity{} <- computeShrinkAction s act]
+data QCDProperty state = QCDProperty
+  { runQCDProperty :: Actions state -> Property
+  , qcdPropertyOptions :: Options state
+  }
 
-      customActionsShrinker :: [(Step state, Annotated state)] -> [[(Step state, Annotated state)]]
-      customActionsShrinker acts =
-        let usedVars = mconcat [getAllVariables a <> getAllVariables (underlyingState s) | (_ := a, s) <- acts]
-            binding (v := _, _) = Some v `Set.member` usedVars
-            -- Remove at most one non-binding action
-            go [] = [[]]
-            go (p : ps)
-              | binding p = map (p :) (go ps)
-              | otherwise = ps : map (p :) (go ps)
-         in go acts
+instance StateModel state => Testable (QCDProperty state) where
+  property QCDProperty{..} =
+    forAllShrink
+      (generateActionsWithOptions qcdPropertyOptions)
+      (shrinkActionsWithOptions qcdPropertyOptions)
+      runQCDProperty
+
+class QCDProp state p | p -> state where
+  qcdProperty :: p -> QCDProperty state
+
+instance QCDProp state (QCDProperty state) where
+  qcdProperty = id
+
+instance Testable p => QCDProp state (Actions state -> p) where
+  qcdProperty p = QCDProperty (property . p) defaultOptions
+
+modifyOptions :: QCDProperty state -> (Options state -> Options state) -> QCDProperty state
+modifyOptions p f =
+  let opts = qcdPropertyOptions p
+   in p{qcdPropertyOptions = f opts}
+
+moreActions :: QCDProp state p => Rational -> p -> QCDProperty state
+moreActions r p =
+  modifyOptions (qcdProperty p) $ \opts -> opts{actionLengthMultiplier = actionLengthMultiplier opts * r}
+
+-- NOTE: indexed on state for forwards compatibility, e.g. when we
+-- want to give an explicit initial state
+data Options state = Options {actionLengthMultiplier :: Rational}
+
+defaultOptions :: Options state
+defaultOptions = Options{actionLengthMultiplier = 1}
+
+-- | Generate arbitrary actions with the `GenActionsOptions`. More flexible than using the type-based
+-- modifiers.
+generateActionsWithOptions :: forall state. StateModel state => Options state -> Gen (Actions state)
+generateActionsWithOptions Options{..} = do
+  (as, rejected) <- arbActions [] [] initialAnnotatedState 1
+  return $ Actions_ rejected (Smart 0 as)
+  where
+    arbActions :: [Step state] -> [String] -> Annotated state -> Int -> Gen ([Step state], [String])
+    arbActions steps rejected s step = sized $ \n -> do
+      let w = round (actionLengthMultiplier * fromIntegral n) `div` 2 + 1
+      continue <- frequency [(1, pure False), (w, pure True)]
+      if continue
+        then do
+          (mact, rej) <- satisfyPrecondition
+          case mact of
+            Just (Some act@ActionWithPolarity{}) -> do
+              let var = mkVar step
+              arbActions
+                ((var := act) : steps)
+                (rej ++ rejected)
+                (computeNextState s act var)
+                (step + 1)
+            Nothing ->
+              return (reverse steps, rejected)
+        else return (reverse steps, rejected)
+      where
+        satisfyPrecondition = sized $ \n -> go n (2 * n) [] -- idea copied from suchThatMaybe
+        go m n rej
+          | m > n = return (Nothing, rej)
+          | otherwise = do
+              a <- resize m $ computeArbitraryAction s
+              case a of
+                Some act ->
+                  if computePrecondition s act
+                    then return (Just (Some act), rej)
+                    else go (m + 1) n (actionName (polarAction act) : rej)
+
+shrinkActionsWithOptions :: forall state. StateModel state => Options state -> Actions state -> [Actions state]
+shrinkActionsWithOptions _ (Actions_ rs as) =
+  map (Actions_ rs) (shrinkSmart (map (prune . map fst) . concatMap customActionsShrinker . shrinkList shrinker . withStates) as)
+  where
+    shrinker :: (Step state, Annotated state) -> [(Step state, Annotated state)]
+    shrinker (v := act, s) = [(unsafeCoerceVar v := act', s) | Some act'@ActionWithPolarity{} <- computeShrinkAction s act]
+
+    customActionsShrinker :: [(Step state, Annotated state)] -> [[(Step state, Annotated state)]]
+    customActionsShrinker acts =
+      let usedVars = mconcat [getAllVariables a <> getAllVariables (underlyingState s) | (_ := a, s) <- acts]
+          binding (v := _, _) = Some v `Set.member` usedVars
+          -- Remove at most one non-binding action
+          go [] = [[]]
+          go (p : ps)
+            | binding p = map (p :) (go ps)
+            | otherwise = ps : map (p :) (go ps)
+       in go acts
 
 -- Running state models
 
@@ -502,47 +543,58 @@ runActions
   :: forall state m e
    . ( StateModel state
      , RunModel state m
-     , e ~ Error state
+     , e ~ Error state m
      , forall a. IsPerformResult e a
      )
   => Actions state
-  -> PropertyM m (Annotated state, Env m)
+  -> PropertyM m (Annotated state, Env)
 runActions (Actions_ rejected (Smart _ actions)) = do
-  (finalState, env) <- runSteps initialAnnotatedState [] actions
+  let bucket = \n -> let (a, b) = go n in show a ++ " - " ++ show b
+        where
+          go n
+            | n < 100 = (d * 10, d * 10 + 9)
+            | otherwise = let (a, b) = go d in (a * 10, b * 10 + 9)
+            where
+              d = div n 10
+  monitor $ tabulate "# of actions" [show $ bucket $ length actions]
+  (finalState, env, names, polars) <- runSteps initialAnnotatedState [] actions
+  monitor $ tabulate "Actions" names
+  monitor $ tabulate "Action polarity" $ map show polars
   unless (null rejected) $
     monitor $
       tabulate "Actions rejected by precondition" rejected
   return (finalState, env)
 
--- | Core function to execute a sequence of `Step` given some initial `Env`ironment
--- and `Annotated` state.
+-- | Core function to execute a sequence of `Step` given some initial `Env`ironment and `Annotated`
+-- state. Return the list of action names and polarities to work around
+-- https://github.com/nick8325/quickcheck/issues/416 causing repeated calls to tabulate being slow.
 runSteps
   :: forall state m e
    . ( StateModel state
      , RunModel state m
-     , e ~ Error state
+     , e ~ Error state m
      , forall a. IsPerformResult e a
      )
   => Annotated state
-  -> Env m
+  -> Env
   -> [Step state]
-  -> PropertyM m (Annotated state, Env m)
-runSteps s env [] = return (s, reverse env)
+  -> PropertyM m (Annotated state, Env, [String], [Polarity])
+runSteps s env [] = return (s, reverse env, [], [])
 runSteps s env ((v := act) : as) = do
   pre $ computePrecondition s act
   ret <- run $ performResultToEither <$> perform (underlyingState s) action (lookUpVar env)
   let name = show polar ++ actionName action
-  monitor $ tabulate "Actions" [name]
-  monitor $ tabulate "Action polarity" [show polar]
   case (polar, ret) of
     (PosPolarity, Left err) ->
       positiveActionFailed err
     (PosPolarity, Right val) -> do
       (s', env') <- positiveActionSucceeded ret val
-      runSteps s' env' as
+      (s'', env'', names, polars) <- runSteps s' env' as
+      pure (s'', env'', name : names, polar : polars)
     (NegPolarity, _) -> do
       (s', env') <- negativeActionResult ret
-      runSteps s' env' as
+      (s'', env'', names, polars) <- runSteps s' env' as
+      pure (s'', env'', name : names, polar : polars)
   where
     polar = polarity act
 
